/*
+-------+-----------------+------+-----------------+-------+-----------------+
|Program| STRCALC.C       |Header| HANSOFT.H       |Library| HANSOFT         |
+-------+-----------------+------+-----------------+-------+----------+------+
|HISTORY                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Turbo C V2.0 version                             | HanSoft          |101193|
| DPX/2 340 version                                | HanSoft          |101193|
+--------------------------------------------------+------------------+------+
|MODIFICATION LOG                                  |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| Fixed some tiny bugs and added ! and @           | HanSoft          |111193|
| Redesigned expression parser; added o, m, l, a, %| HanSoft          |121193|
| Made _StrCalcErr, _ToS and _CStack static vars   | HanSoft          |151193|
| Renamed _pop() to pop(), _push() to push()       | HanSoft          |280294|
| Made pop() and push() static; added ANSI_C supp. | HanSoft          |280294|
| Made long datatype; added variables              | HanSoft          |220895|
| Removed EasyC                                    | HanSoft          |100401|
+--------------------------------------------------+------------------+------+
|REMARKS                                           |    Programmer    | Date |
+--------------------------------------------------+------------------+------+
| None                                             |                  |      |
+--------------------------------------------------+------------------+------+

  Copyright (C) 1991,2002 J.L. Bezemer

  This file is part of BDDE

  BDDE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  
Interprets a Reverse Polish Notation Expression in "expression" and assigns
the internal variables to Vals. Returns an integer (Top of Stack) on succes or
LONG_MIN when an error occurs.

These operators are available:

Stack operators:

  x y s (swaps x and y on the stack)
  x c   (copies x on the stack)
  x d   (drops x from the stack)
  x y o (duplicates x on the stack)
  x y ! (stores x in variable y)
  x @   (puts contents of variable x on stack)

Mathematical operators:

  x y + (adds x and y and puts the result on the stack)
  x y - (subtracts x and y and puts the result on the stack)
  x y * (multiplies x and y and puts the result on the stack)
  x y / (divides x and y and puts the result on the stack)
  x y % (divides x and y and puts the remainder on the stack)
  x y l (takes x and y from stack and returns the smallest on the stack)
  x y m (takes x and y from stack and returns the largest on the stack)
  x a   (takes the absolute of x and returns it on the stack)
  x n   (negates x and puts the result on the stack)

Binary operators:

  x y | (ORs x and y and puts the result on the stack)
  x y & (ANDs x and y and puts the result on the stack)
  x y ^ (XORs x and y and puts the result on the stack)
  x y > (shifts x right y and puts the result on the stack)
  x y < (shifts x left y and puts the result on the stack)
  x ~   (NOTs x and puts the result on the stack)

Prototype: long Strcalc (char *expression, unsigned NoVars, ...)
Returns  : long on success; LONG_MIN on error
Memory   : None
Example  :


#include <stdio.h>
#include <hansoft.h>

void main (void)

{
  printf ("%ld\n", Strcalc ("20 0 @ >", 1, 1L));
  printf ("%ld\n", Strcalc ("20 1 + n", 0));
  printf ("%ld\n", Strcalc ("20 2 / ", 0));
  printf ("%ld\n", Strcalc ("0 @ 20 1 @ * s /", 2, 4L, 2L));
}


Input    : None
Output   : 10
           -21
           10
           10
Related  : atoi(), atol()
*/


#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>

#ifdef USRLIBHANSOFT
#include <hansoft.h>
#else
#include "hansoft.h"
#endif

#define MAXOPD  32
#define MAXVARS 16

static long Stack [MAXOPD];
static long Vars  [MAXVARS];
static unsigned ToS, CalcErr;


/*
This routine pushes an operand on the stack. If the stack is full it sets
CalcErr
*/

#ifndef ARCHAIC
  static void push (long val)
#else
  static void push (val) long val;
#endif

{
  if (ToS < MAXOPD) Stack [ToS++] = val;
  else CalcErr = TRUE;
}


/*
This routine pops an operand from the stack. If the stack is empty it sets
CalcErr
*/

#ifndef ARCHAIC
  static void pop (long *val)
#else
  static void pop (val) long* val;
#endif

{
  if (ToS > 0) *val = Stack [--ToS];
  else CalcErr = TRUE;
}


#ifndef ARCHAIC
  long Strcalc (char *expression, unsigned NoVars, ...)
#else
  long Strcalc (expression, NoVars) char *expression; unsigned NoVars;
#endif

{
  long a, b;                           /* internal stack variables */
  char *p;                             /* pointer to next token */
  va_list Vals;                        /* list of preloaded variables */

  ToS = CalcErr = 0;                   /* reset stackpointer and errorflag */
                                       /* preload variables */
  va_start (Vals, NoVars);
  for (a = 0; a < (NoVars < MAXVARS ? NoVars : MAXVARS); a++)
      Vars [(unsigned) a] = va_arg (Vals, long);
  va_end (Vals);

/*
This is the main calculation engine. All operators can be reduced to pop() and
push() calls. It executes until there are no more operators left.
*/

  for ( p = strtok (expression, " ");
       (p) && (! CalcErr);
        p = strtok (NULL, " ")
      )

      switch (p[0])
      {
        case ('c'): pop (&a); push (a); push (a); break;
        case ('s'): pop (&a); pop (&b); push (a); push (b); break;
        case ('d'): pop (&a); break;
        case ('o'): pop (&b); pop (&a); push (a); push (b); push (a);
                    break;
        case ('!'): pop (&a); pop (&b);
                    if ((a < MAXVARS) && (a >= 0L))
                       Vars [(unsigned) a] = b;
                    else CalcErr = TRUE; break;
        case ('@'): pop (&a);
                    if ((a < MAXVARS) && (a >= 0L))
                       push (Vars [(unsigned) a]);
                    else CalcErr = TRUE; break;
        case ('+'): pop (&b); pop (&a); push (a + b); break;
        case ('-'): if ((strlen (p) > 1) && (isdigit ((int) p[1])))
                       push (atol (p));
                    else
                       { 
                         pop (&b); pop (&a); push (a - b);
                       }
                    break;
        case ('*'): pop (&b); pop (&a); push (a * b); break;
        case ('/'): pop (&b); pop (&a);
                    if (b != 0) push (a / b); else CalcErr = TRUE;
                    break;
        case ('%'): pop (&b); pop (&a);
                    if (b != 0) push (a % b); else CalcErr = TRUE;
                    break;
        case ('l'): pop (&b); pop (&a); push ((a < b ? a : b)); break;
        case ('m'): pop (&b); pop (&a); push ((a > b ? a : b)); break;
        case ('a'): pop (&a); push ((a < 0 ? -a : a)); break;
        case ('n'): pop (&a); push (-a); break;
        case ('|'): pop (&b); pop (&a); push (a | b); break;
        case ('&'): pop (&b); pop (&a); push (a & b); break;
        case ('^'): pop (&b); pop (&a); push (a ^ b); break;
        case ('>'): pop (&b); pop (&a); push (a >> b); break;
        case ('<'): pop (&b); pop (&a); push (a << b); break;
        case ('~'): pop (&a); push (~a); break;
        default: if (isdigit ((int) p[0])) push (atol (p));
                 else CalcErr = TRUE; break;
      }

   pop (&a);
   if (CalcErr) return (LONG_MIN); else return (a);
}
